import json
import os
import sys
import tempfile
from pathlib import Path

import yaml

from album.core.api.model.configuration import IConfiguration
from album.core.api.model.environment import IEnvironment
from album.core.api.model.link import Link
from album.core.model.default_values import DefaultValues
from album.core.utils import subcommand
from album.core.utils.operations.file_operations import construct_cache_link_target, remove_link
from album.core.utils.subcommand import SubProcessError
from album.runner import album_logging
from album.runner.album_logging import debug_settings

module_logger = album_logging.get_active_logger


class CondaManager:
    """Class for handling conda environments.

    The conda class manages the environments a solution is supposed to run in. It provides all features necessary for
    environment creation, deletion, dependency installation, etc.

    Notes:
        An installed \"conda\" program must be available and callable via commandline or powershell.

    """

    def __init__(self, configuration: IConfiguration):
        self._configuration = configuration
        self._conda_executable = self._configuration.conda_executable()

    def _get_install_environment_executable(self):
        return self._conda_executable

    def get_environment_list(self):
        """Returns the available album conda environments."""
        if Path(self._get_base_environment_target_path()).exists():
            return sorted(self._get_immediate_subdirectories(self._get_base_environment_target_path()))
        else:
            return []

    @staticmethod
    def _get_immediate_subdirectories(a_dir: Path):
        return [a_dir.joinpath(name).resolve() for name in os.listdir(a_dir)
                if os.path.isdir(os.path.join(a_dir, name))]

    def _get_base_environment_target_path(self):
        """Gets the first of the paths the conda installation uses to manage its environments."""
        return Path(self._configuration.lnk_path()).joinpath(DefaultValues.lnk_env_prefix.value)

    def environment_exists(self, environment_name):
        """Checks whether an environment already exists or not.

        Args:
            environment_name:
                The name of the environment in which a solution will run.

        Returns:
            True when environment exists else false.
        """
        environment_list = self.get_environment_list()
        path_expected = self._environment_name_to_path(environment_name, create=False)

        return True if (path_expected and path_expected.resolve() in environment_list and os.listdir(path_expected)) else False

    def _environment_name_to_path(self, environment_name, create=True):
        path = os.path.normpath(self._configuration.cache_path_envs().joinpath(environment_name))
        target = construct_cache_link_target(self._configuration.lnk_path(), link=path,
                                             target=DefaultValues.lnk_env_prefix.value, create=create)
        if target:
            return Link(target).set_link(path)
        else:
            return None

    def get_environment_path(self, environment_name: str, create: bool = True) -> Link:
        """Gets the environment path for a given environment

        Args:
            environment_name:
                The environment to get the path for.

        Returns:
            None or the path
        """
        environment_list = self.get_environment_list()
        path_expected = self._environment_name_to_path(environment_name, create)

        if path_expected.resolve() in environment_list:
            return path_expected
        raise LookupError('Could not find environment %s.' % environment_name)

    def get_active_environment_name(self):
        """Returns the environment from the active album."""
        conda_list = self.get_info()
        return conda_list["active_prefix_name"]

    def get_active_environment_path(self):
        """Returns the environment form the active album."""
        conda_list = self.get_info()
        path = conda_list["active_prefix"]
        link = construct_cache_link_target(self._configuration.lnk_path(), link=path,
                                             target=DefaultValues.lnk_env_prefix.value, create=False)
        if link:
            return link
        else:
            return Link(path)

    def remove_environment(self, environment_name) -> bool:
        """Removes an environment given its name. Does nothing when environment does not exist.

        Args:
            environment_name:
                The name of the environment to remove

        Returns:
            True, when removal succeeded, else False

        """
        if self.get_active_environment_name() == environment_name:
            module_logger().warning("Cannot remove active environment! Skipping...")
            return False

        if not self.environment_exists(environment_name):
            module_logger().warning("Environment does not exist! Skipping...")
            return False

        path = self.get_environment_path(environment_name, create=False)

        try:
            subprocess_args = [
                self._get_install_environment_executable(), 'env', 'remove', '-y', '-q', '-p', os.path.normpath(path)
            ]

            subcommand.run(subprocess_args, log_output=False)
        except SubProcessError:
            module_logger().debug("Can't delete environment via command line call, deleting the folder next...")
        # try to remove file content if any but don't fail:
        remove_link(path)
        return True

    def get_info(self):
        """Get the info of the conda installation on the corresponding system.

        Returns:
            dictionary corresponding to conda info.
        """
        subprocess_args = [
            self._get_install_environment_executable(), 'info', '--json'
        ]
        output = subcommand.check_output(subprocess_args)
        return json.loads(output)

    def list_environment(self, environment_path):
        """Lists all available conda installation in the given environment.

        Args:
            environment_path:
                The prefix of the environment to list.

        Returns:
            dictionary containing the available packages in the given conda environment.
        """
        subprocess_args = [self._get_install_environment_executable(), 'list', '--json', '--prefix', str(environment_path)]
        output = subcommand.check_output(subprocess_args)
        return json.loads(output)

    def create_environment_from_file(self, yaml_path, environment_name, album_api_version=None):
        """Creates a conda environment given a path to a yaml file and its name.

        Args:
            yaml_path:
                The path to the file.
            environment_name:
                The name of the environment. Must be the same as specified in the yml file.

        Raises:
            NameError:
                When the file has the wrong format according to its extension.
            ValueError:
                When the file is unreadable or empty.
            RuntimeError:
                When the environment could not be created due to whatever reasons.

        """
        if self.environment_exists(environment_name):
            self.remove_environment(environment_name)

        if not (str(yaml_path).endswith(".yml") or str(yaml_path).endswith(".yaml")):
            raise NameError("File needs to be a yml or yaml file!")

        yaml_path = Path(yaml_path)

        if not (yaml_path.is_file() and yaml_path.stat().st_size > 0):
            raise ValueError("File not a valid yml file!")

        if not album_api_version:
            album_api_version = DefaultValues.runner_api_packet_version.value
        if not DefaultValues.runner_api_packet_version.value:
            album_api_version = None

        with open(yaml_path, 'r') as f:
            content = yaml.safe_load(f)

        self._append_framework_to_yml(content, album_api_version)

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.yml') as env_file:
            env_file.write(yaml.safe_dump(content))

        env_prefix = os.path.normpath(self._environment_name_to_path(environment_name))

        # TODO in debug mode, use -v to display the installation process
        subprocess_args = [self._get_install_environment_executable(), 'env', 'create', '-q', '--force', '-f',
                           os.path.normpath(env_file.name), '-p', env_prefix]

        try:
            subcommand.run(subprocess_args, log_output=True)
        except SubProcessError as e:
            # cleanup after failed installation
            if self.environment_exists(environment_name):
                module_logger().debug('Cleanup failed installation...')
                self.remove_environment(environment_name)
            raise RuntimeError("Command failed due to reasons above!") from e
        # finally:
        #     os.remove(env_file.name)

    @staticmethod
    def _append_framework_to_yml(content, album_api_version):
        framework = '%s==%s' % (DefaultValues.runner_api_packet_name.value, album_api_version)
        if not 'dependencies' in content or not content['dependencies']:
            content['dependencies'] = []
        pip_dict = None
        found_pip_dep = False
        for dep in content['dependencies']:
            if isinstance(dep, dict):
                if 'pip' in dep:
                    pip_dict = dep
            if dep == 'pip' or str(dep).startswith('pip='):
                found_pip_dep = True
        if not found_pip_dep:
            content['dependencies'].append(DefaultValues.runner_pip_version.value)
        if not pip_dict:
            pip_dict = {'pip': []}
            content['dependencies'].append(pip_dict)
        pip_dict['pip'].append(framework)
        return content

    def create_environment(self, environment_name, album_api_version=None, force=False):
        """Creates a conda environment with python (latest version) installed.

        Args:
            environment_name:
                The desired environment name.
            force:
                If True, force creates the environment by deleting the old one.
        Raises:
            RuntimeError:
                When the environment could not be created due to whatever reasons.

        """
        env_exists = self.environment_exists(environment_name)
        if force and env_exists:
            self.remove_environment(environment_name)
        else:
            if env_exists:
                raise EnvironmentError("Environment with name %s already exists!" % environment_name)

        env_prefix = os.path.normpath(self._environment_name_to_path(environment_name))

        if not album_api_version:
            album_api_version = DefaultValues.runner_api_packet_version.value
        if not DefaultValues.runner_api_packet_version.value:
            album_api_version = None

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.yml') as env_file:

            env_file.write(
                "channels:\n" +
                "  - defaults\n" +
                "dependencies:\n" +
                "  - python=3.8\n" +
                "  - pip\n" +
                "  - pip:\n" +
                "    - %s==%s\n" % (DefaultValues.runner_api_packet_name.value, album_api_version)
            )

        subprocess_args = [self._get_install_environment_executable(), 'env', 'create', '--force', '-q',
                           '--file', env_file.name, '-p', env_prefix]

        try:
            subcommand.run(subprocess_args, log_output=True)
        except RuntimeError as e:
            # cleanup after failed installation
            if self.environment_exists(environment_name):
                module_logger().debug('Cleanup failed environment creation...')
                self.remove_environment(environment_name)
            raise RuntimeError("Command failed due to reasons above!") from e
        finally:
            os.remove(env_file.name)

    def run_script(self, environment_path, script_full_path, pipe_output=True):
        """Runs a script in the given environment.

        Args:
            environment_path:
                The prefix path of the environment to install the package to.
            script_full_path:
                The full path to the script to run.
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.

        """
        if sys.platform == 'win32' or sys.platform == 'cygwin':
            # NOTE: WHEN USING 'CONDA RUN' THE CORRECT ENVIRONMENT GETS TEMPORARY ACTIVATED,
            # BUT THE PATH POINTS TO THE WRONG PYTHON (conda base folder python) BECAUSE THE CONDA BASE PATH
            # COMES FIRST IN ENVIRONMENT VARIABLE "%PATH%". THUS, FULL PATH IS NECESSARY TO CALL
            # THE CORRECT PYTHON OR PIP! ToDo: keep track of this!
            subprocess_args = [
                self._conda_executable, 'run', '--no-capture-output', '--prefix',
                os.path.normpath(environment_path),
                os.path.normpath(Path(environment_path).joinpath('python')),
                os.path.normpath(script_full_path)
            ]
        else:
            subprocess_args = [
                self._conda_executable, 'run', '--no-capture-output', '--prefix',
                os.path.normpath(environment_path), 'python',
                os.path.normpath(script_full_path)
            ]
        subcommand.run(subprocess_args, pipe_output=pipe_output)

    def set_environment_path(self, environment: IEnvironment):
        path = self.get_environment_path(environment.name())
        module_logger().debug('Set environment path to %s...' % path)
        environment.set_path(path)

    def is_installed(self, environment_path: str, package_name, min_package_version=None):
        """Checks if package is installed in a certain version."""
        conda_list = self.list_environment(environment_path)

        for package in conda_list:
            if package["name"] == package_name:
                if min_package_version:
                    if package["version"] == min_package_version:
                        module_logger().debug('Package %s:%s is installed...' % (package_name, min_package_version))
                        return True
                    if package["version"] < min_package_version:
                        module_logger().debug('Package %s:%s is installed. Requirements not set! Reinstalling...'
                                              % (package_name, package["version"]))
                        return False
                    if package["version"] > min_package_version:
                        module_logger().debug('Package %s:%s is installed. Version should be compatible...'
                                              % (package_name, package["version"]))
                        return True
                else:
                    module_logger().debug('Package %s:%s is installed...' % (package_name, package["version"]))
                    return True

        return False

    def run_scripts(self, environment: IEnvironment, scripts, pipe_output=True):
        """Runs the solution in the target environment

        Args:
            scripts:
                List of he scripts calling the solution(s)
            environment:
                The virtual environment used to run the scripts
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.
        """
        if not environment.path():
            raise EnvironmentError('Could not find environment %s. Is the solution installed?' % environment.name())

        module_logger().debug('run_in_environment: %s...' % str(environment.path()))

        # first write scripts to disc, create meta-script to execute them in the order of the list
        if len(scripts) > 1:
            script = ""
            for s in scripts:
                fp_step = tempfile.NamedTemporaryFile(mode='w+', delete=False)
                fp_step.write(s)
                fp_step.flush()
                os.fsync(fp_step)
                script += "\nexec(open(r\'%s\').read())\n" % fp_step.name
                fp_step.close()
        else:
            script = scripts[0]

        self.write_and_run_script(environment, script, pipe_output=pipe_output)

    def write_and_run_script(self, environment, script, pipe_output):
        """Use an environment path and a temporary file to store the script
        Args:

            environment:
                The virtual environment used to run the scripts
            script:
                The virtual environment used to run the scripts
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.
        """
        if debug_settings():
            fp = open(str(DefaultValues.app_cache_dir.value.joinpath('album_test.py')), 'w')
            module_logger().debug(
                "Executable file in: %s..." % str(DefaultValues.app_cache_dir.value.joinpath('album_test.py')))
        else:
            fp = tempfile.NamedTemporaryFile(mode='w+', delete=False)
            module_logger().debug('Executable file in: %s...' % fp.name)
        fp.write(script)
        fp.flush()
        os.fsync(fp)
        fp.close()
        self.run_script(environment.path(), fp.name, pipe_output=pipe_output)
        Path(fp.name).unlink()

    def create_or_update_env(self, environment: IEnvironment, album_api_version: str = None):
        """Creates or updates the environment"""
        if self.environment_exists(environment.name()):
            self.update(environment)
        else:
            self.create(environment, album_api_version)

    def update(self, environment: IEnvironment):
        """Updates the environment"""
        module_logger().debug('Skip installing environment %s...' % environment.name())
        pass  # ToDo: implement and change log message

    def create(self, environment: IEnvironment, album_api_version=None):
        """Creates environment a solution runs in."""
        if environment.yaml_file():
            self.create_environment_from_file(environment.yaml_file(), environment.name(), album_api_version)
        else:
            module_logger().warning("No yaml file specified. Creating Environment without dependencies!")
            self.create_environment(environment.name(), album_api_version)

    def install(self, environment: IEnvironment, album_api_version=None):
        """Creates or updates an an environment and installs album in the target environment."""
        self.create_or_update_env(environment, album_api_version)
        self.set_environment_path(environment)



import os
from enum import Enum
from pathlib import Path

from appdirs import user_data_dir, user_cache_dir, user_config_dir


class DefaultValues(Enum):
    """Add an entry here to initialize default attributes for a album framework installation instance."""

    # runner
    runner_api_packet_version = '0.3.1'      # set to None to use with url
    runner_api_packet_name = 'album-runner'  # can also point to zip/url like: https://gitlab.com/album-app/album-runner/-/archive/main/album-runner-main.zip
    runner_pip_version = 'pip=21.0'

    # templates
    catalog_template_url = 'https://gitlab.com/album-app/catalogs/templates'  # base URL of available catalog templates

    # catalog
    local_catalog_name = 'catalog_local'                                   # the default name of the local catalog (always configured)
    _catalog_url = 'https://gitlab.com/album-app/catalogs/default'
    _catalog_branch = "main"
    default_catalog_src = os.getenv('ALBUM_DEFAULT_CATALOG', _catalog_url) # default catalog, either catalog_url or env. variable
    default_catalog_src_branch = os.getenv(
        'ALBUM_DEFAULT_CATALOG_BRANCH', _catalog_branch
    )                                                                      # default catalog branch either _catalog_branch or env. variable
    default_catalog_name = "default"                                       # default catalog, either catalog_url or env. variable
    catalog_collection_name = 'album_collection'                           # the default name of the Collection
    catalog_collection_db_name = 'catalog_collection.db'                   # the default name of the Collection DB
    catalog_collection_json_name = 'catalog_collection.json'               # the default name of the Collection JSON
    catalog_index_file_name = 'album_catalog_index.db'                     # the default index file name of the catalog_index
    catalog_index_metafile_json = 'album_catalog_index.json'               # the default meta file name of the catalog_index
    catalog_solution_list_file_name = 'album_solution_list.json'           # the default file name for exporting the list of solutions of a catalog
    catalog_folder_prefix = 'catalogs'                                     # base folder prefix where all not local catalogs live
    link_folder_prefix = 'lnk'                                             # base folder prefix where all internal link destinations live
    cache_path_solution_prefix = "solutions"                               # base folder prefix where solutions live
    cache_path_app_prefix = "apps"                                         # base folder prefix where app solutions live
    cache_path_data_prefix = "data"                                        # base folder prefix where data lives
    cache_path_download_prefix = "downloads"                               # base folder prefix where downloads live
    cache_path_tmp_prefix = "tmp"                                          # base folder prefix where temporary files live
    cache_path_envs_prefix = "envs"                                        # base folder prefix where environments live in

    # solutions
    solution_default_name = "solution.py"                                  # default name how solution.py files are called
    changelog_default_name = "CHANGELOG.md"                                # default name of the changelog file

    # lnk folder prefixes
    lnk_data_prefix = 'data'
    lnk_app_prefix = 'app'
    lnk_package_prefix = 'pck'
    lnk_internal_cache_prefix = 'icache'
    lnk_user_cache_prefix = 'ucache'
    lnk_env_prefix = 'env'

    # environment
    default_environment = "album"            # default environment name the album framework operates from

    # album
    app_data_dir = Path(user_data_dir("album"))      # base data path
    app_cache_dir = Path(user_cache_dir("album"))    # base cache path
    app_config_dir = Path(user_config_dir("album"))  # base configuration path

    # conda
    conda_default_executable = "conda"                                   # default conda executable
    conda_path = os.getenv('ALBUM_CONDA_PATH', conda_default_executable)  # default conda path, either env. var or conda

    # server
    server_port = 5476  # default port used to launch server
    server_host = "127.0.0.1"  # default host used to launch server. Set to 0.0.0.0 when used in docker



from datetime import datetime
from pathlib import Path

from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_updates import ISolutionChange, ChangeType
from album.core.api.model.link import Link
from album.core.model.catalog import get_solution_src
from album.core.model.collection_index import CollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import copy_folder, copy, unzip_archive, construct_cache_link_target
from album.core.utils.operations.resolve_operations import dict_to_coordinates, get_zip_name
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.core.utils.operations.url_operations import download_resource
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution

module_logger = album_logging.get_active_logger


class SolutionHandler(ISolutionHandler):
    """Handles everything inside the Collection responsible for a solution entry.

    Is NOT responsible for resolving paths as this is part of a catalog.
    """

    def __init__(self, album: IAlbumController):
        self.album = album

    def add_or_replace(self, catalog: ICatalog, active_solution: ISolution, path):
        deploy_dict = get_deploy_dict(active_solution)
        self._get_collection_index().add_or_replace_solution(
            catalog.catalog_id(),
            active_solution.coordinates(),
            deploy_dict
        )
        # get the install location
        install_location = self.get_solution_path(catalog, dict_to_coordinates(deploy_dict))

        if Path(path).is_dir():
            copy_folder(path, install_location, copy_root_folder=False)
        else:
            copy(path, install_location.joinpath(DefaultValues.solution_default_name.value))

    def add_to_local_catalog(self, active_solution: ISolution, path):
        self.add_or_replace(self.album.catalogs().get_local_catalog(), active_solution, path)

    def set_parent(self, catalog_parent: ICatalog, catalog_child: ICatalog, coordinates_parent: ICoordinates,
                   coordinates_child: ICoordinates):

        # retrieve parent entry
        parent_entry = self._get_collection_index().get_solution_by_catalog_grp_name_version(
            catalog_parent.catalog_id(), coordinates_parent, close=False
        )
        # retrieve child entry
        child_entry = self._get_collection_index().get_solution_by_catalog_grp_name_version(
            catalog_child.catalog_id(), coordinates_child, close=False
        )

        self._get_collection_index().insert_collection_collection(
            parent_entry.internal()["collection_id"],
            child_entry.internal()["collection_id"],
            catalog_parent.catalog_id(),
            catalog_child.catalog_id()
        )

    def remove_parent(self, catalog: ICatalog, coordinates: ICoordinates):
        entry = self._get_collection_index().get_solution_by_catalog_grp_name_version(
            catalog.catalog_id(), coordinates, close=False
        )
        self._get_collection_index().remove_parent(
            entry.internal()["collection_id"]
        )

    def remove_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        self._get_collection_index().remove_solution(catalog.catalog_id(), coordinates)

    def update_solution(self, catalog: ICatalog, coordinates: ICoordinates, attrs: dict):
        self._get_collection_index().update_solution(
            catalog.catalog_id(),
            coordinates,
            attrs,
            CollectionIndex.get_collection_column_keys()
        )

    def apply_change(self, catalog: ICatalog, change: ISolutionChange):
        # FIXME handle other tables (tags etc)
        if change.change_type() is ChangeType.ADDED:
            self._get_collection_index().add_or_replace_solution(
                catalog.catalog_id(),
                change.coordinates(),
                catalog.index().get_solution_by_coordinates(change.coordinates())
            )

        elif change.change_type is ChangeType.REMOVED:
            self.remove_solution(catalog, change.coordinates())

        elif change.change_type is ChangeType.CHANGED:
            self.remove_solution(catalog, change.coordinates())
            self._get_collection_index().add_or_replace_solution(
                catalog.catalog_id(),
                change.coordinates(),
                catalog.index().get_solution_by_coordinates(change.coordinates())
            )

    def set_installed(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(
            catalog,
            coordinates,
            {"installed": 1, "installation_unfinished": 0, "install_date": datetime.now().isoformat()}
        )

    def set_uninstalled(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(catalog, coordinates, {"installed": 0, "installation_unfinished": 0})

    def set_installation_unfinished(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(catalog, coordinates, {"installed": 0, "installation_unfinished": 1})

    def is_installed(self, catalog: ICatalog, coordinates: ICoordinates) -> bool:
        try:
            return self._get_collection_index().is_installed(catalog.catalog_id(), coordinates)
        except LookupError:
            return False

    def get_solution_path(self, catalog: ICatalog, coordinates: ICoordinates):
        base_link = catalog.path().joinpath(self.album.configuration().get_solution_path_suffix(coordinates))
        link_target = construct_cache_link_target(self.album.configuration().lnk_path(), base_link,
                                                  DefaultValues.lnk_package_prefix.value)
        return Link(link_target).set_link(link=base_link)

    def get_solution_file(self, catalog: ICatalog, coordinates: ICoordinates):
        return self.get_solution_path(catalog, coordinates).joinpath(DefaultValues.solution_default_name.value)

    def get_solution_zip(self, catalog: ICatalog, coordinates: ICoordinates):
        return self.get_solution_path(catalog, coordinates).joinpath(get_zip_name(coordinates))

    def get_solution_zip_suffix(self, coordinates: ICoordinates):
        return Path("").joinpath(
            self.album.configuration().get_solution_path_suffix(coordinates),
            get_zip_name(coordinates)
        )

    def retrieve_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        if catalog.is_cache():  # no src to download form or src to copy from
            raise RuntimeError("Cannot download from a cache catalog!")

        elif catalog.is_local():  # src to copy from
            src_path = Path(catalog.src()).joinpath(self.get_solution_zip_suffix(coordinates))
            solution_zip_file = self.get_solution_zip(catalog, coordinates)
            copy(src_path, solution_zip_file)

        else:  # src to download from
            url = get_solution_src(catalog.src(), coordinates, catalog.branch_name())
            solution_zip_file = self.get_solution_zip(catalog, coordinates)
            download_resource(url, solution_zip_file)

        solution_zip_path = unzip_archive(solution_zip_file)
        solution_path = solution_zip_path.joinpath(DefaultValues.solution_default_name.value)

        return solution_path

    def set_cache_paths(self, solution: ISolution, catalog: ICatalog):
        # Note: cache paths need the catalog the solution lives in - otherwise there might be problems with solutions
        # of different catalogs doing similar operations (e.g. downloads) as they might share the same cache path.

        catalog_name = catalog.name()
        path_suffix = Path("").joinpath(solution.coordinates().group(), solution.coordinates().name(),
                                        solution.coordinates().version())

        solution.installation().set_package_path(self.get_solution_path(catalog, solution.coordinates()))

        self._set_cache_path(
            solution.installation().set_data_path,
            self.album.configuration().cache_path_data().joinpath(str(catalog_name), path_suffix),
            DefaultValues.lnk_data_prefix.value
        )
        self._set_cache_path(
            solution.installation().set_app_path,
            self.album.configuration().cache_path_app().joinpath(str(catalog_name), path_suffix),
            DefaultValues.lnk_app_prefix.value
        )

        self._set_cache_path(
            solution.installation().set_internal_cache_path,
            self.album.configuration().cache_path_tmp_internal().joinpath(str(catalog_name), path_suffix),
            DefaultValues.lnk_internal_cache_prefix.value
        )
        self._set_cache_path(
            solution.installation().set_user_cache_path,
            self.album.configuration().cache_path_tmp_user().joinpath(str(catalog_name), path_suffix),
            DefaultValues.lnk_user_cache_prefix.value
        )

    def _set_cache_path(self, path_method, link, link_target_prefix):
        link_target = construct_cache_link_target(self.album.configuration().lnk_path(), link, link_target_prefix)
        path_method(Link(link_target).set_link(link))

    def _get_collection_index(self):
        return self.album.collection_manager().get_collection_index()



from queue import Queue

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.test_manager import ITestManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.runner import album_logging
from album.runner.core.model.script_creator import ScriptCreatorTest

module_logger = album_logging.get_active_logger


class TestManager(ITestManager):

    def __init__(self, album: IAlbumController):
        self.album = album

    def test(self, resolve_result: ICollectionSolution, args=None):
        if args is None:
            args = [""]

        solution = resolve_result.loaded_solution()

        if solution.setup().pre_test and callable(solution.setup().pre_test) \
                and solution.setup().test and callable(solution.setup().test):
            queue = Queue()
            script_test_creator = ScriptCreatorTest()

            # do not run queue immediately
            self.album.script_manager().build_queue(resolve_result, queue, script_test_creator, False, args)

            # runs the queue
            self.album.script_manager().run_queue(queue)

            module_logger().info('Ran test routine for \"%s\"!' % resolve_result.loaded_solution().coordinates().name())
        else:
            module_logger().warning(
                'No \"test\" routine configured for solution \"%s\"! Skipping...' % solution.coordinates().name()
            )



import os
import platform
import shutil
import sys
from pathlib import Path

from album.core.api.model.configuration import IConfiguration
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import create_paths_recursively, force_remove, get_dict_from_json
from album.runner import album_logging
from album.runner.core.model.coordinates import Coordinates

module_logger = album_logging.get_active_logger


class Configuration(IConfiguration):

    def __init__(self):
        self._is_setup = False
        self._base_cache_path = None
        self._conda_executable = None
        self._mamba_executable = None
        self._cache_path_app = None
        self._cache_path_data = None
        self._cache_path_download = None
        self._cache_path_tmp_internal = None
        self._cache_path_tmp_user = None
        self._cache_path_envs = None
        self._catalog_collection_path = None
        self._lnk_path = None

    def base_cache_path(self):
        return self._base_cache_path

    def conda_executable(self):
        return self._conda_executable

    def mamba_executable(self):
        return self._mamba_executable

    def cache_path_app(self):
        return self._cache_path_app

    def cache_path_data(self):
        return self._cache_path_data

    def cache_path_download(self):
        return self._cache_path_download

    def cache_path_tmp_internal(self):
        return self._cache_path_tmp_internal

    def cache_path_tmp_user(self):
        return self._cache_path_tmp_user

    def cache_path_envs(self):
        return self._cache_path_envs

    def lnk_path(self):
        return self._lnk_path

    def is_setup(self):
        return self._is_setup

    def setup(self, base_cache_path=None):
        if self._is_setup:
            raise RuntimeError("Configuration::setup was already called and should not be called twice.")
        self._is_setup = True
        # base root path where everything lives
        self._base_cache_path = Path(os.getenv('ALBUM_BASE_CACHE_PATH', DefaultValues.app_data_dir.value))
        if base_cache_path:
            self._base_cache_path = Path(base_cache_path)

        # conda executable
        conda_path = DefaultValues.conda_path.value
        operation_system = platform.system().lower()
        if conda_path is not DefaultValues.conda_default_executable.value:
            self._conda_executable = self._build_conda_executable(conda_path)
        else:
            self._conda_executable = conda_path
            if 'windows' in operation_system:
                self._conda_executable = shutil.which(self._conda_executable)

        self._mamba_executable = 'mamba'
        if 'windows' in operation_system:
            self._mamba_executable = shutil.which(self._mamba_executable)
        self._cache_path_tmp_internal = self._base_cache_path.joinpath(DefaultValues.cache_path_solution_prefix.value)
        self._cache_path_app = self._base_cache_path.joinpath(DefaultValues.cache_path_app_prefix.value)
        self._cache_path_data = self._base_cache_path.joinpath(DefaultValues.cache_path_data_prefix.value)
        self._cache_path_download = self._base_cache_path.joinpath(DefaultValues.cache_path_download_prefix.value)
        self._cache_path_tmp_user = self._base_cache_path.joinpath(DefaultValues.cache_path_tmp_prefix.value)
        self._cache_path_envs = self._base_cache_path.joinpath(DefaultValues.cache_path_envs_prefix.value)
        self._catalog_collection_path = self._base_cache_path.joinpath(DefaultValues.catalog_folder_prefix.value)
        self._lnk_path = self._base_cache_path.joinpath(DefaultValues.link_folder_prefix.value)
        create_paths_recursively(
            [
                self._cache_path_tmp_internal,
                self._cache_path_app,
                self._cache_path_data,
                self._cache_path_download,
                self._cache_path_tmp_user,
                self._cache_path_envs,
                self._catalog_collection_path
            ]
        )

        self._empty_tmp()

    @staticmethod
    def _build_conda_executable(conda_path):
        operation_system = sys.platform
        if operation_system == 'linux' or operation_system == 'darwin':
            return str(Path(conda_path).joinpath("bin", "conda"))
        else:
            return str(Path(conda_path).joinpath("Scripts", "conda.exe"))

    def get_solution_path_suffix(self, coordinates: Coordinates) -> Path:
        return Path("").joinpath(
            DefaultValues.cache_path_solution_prefix.value,
            coordinates.group(),
            coordinates.name(),
            coordinates.version()
        )

    def get_cache_path_catalog(self, catalog_name):
        return self._base_cache_path.joinpath(DefaultValues.catalog_folder_prefix.value, catalog_name)

    def get_catalog_collection_path(self):
        collection_db_path = Path(self._catalog_collection_path).joinpath(
            DefaultValues.catalog_collection_db_name.value)
        return collection_db_path

    def get_catalog_collection_meta_dict(self):
        """Returns the metadata of the collection as a dict."""
        catalog_collection_json = self.get_catalog_collection_meta_path()
        if not catalog_collection_json.exists():
            return None
        catalog_collection_dict = get_dict_from_json(catalog_collection_json)
        return catalog_collection_dict

    def get_catalog_collection_meta_path(self):
        return self._catalog_collection_path.parent.joinpath(
            DefaultValues.catalog_collection_json_name.value
        )

    def get_initial_catalogs(self):
        return {
            DefaultValues.local_catalog_name.value: self.get_cache_path_catalog(DefaultValues.local_catalog_name.value),
            DefaultValues.default_catalog_name.value: DefaultValues.default_catalog_src.value
        }

    def get_initial_catalogs_branch_name(self):
        return {
            DefaultValues.local_catalog_name.value: "main",
            DefaultValues.default_catalog_name.value: DefaultValues.default_catalog_src_branch.value
        }

    def _empty_tmp(self):
        """Removes the content of the tmp folder"""
        force_remove(self._cache_path_tmp_user)



from pathlib import Path

from album.core.api.controller.clone_manager import ICloneManager
from album.core.api.controller.controller import IAlbumController
from album.core.model.default_values import DefaultValues
from album.core.utils.operations import url_operations, file_operations
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class CloneManager(ICloneManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def clone(self, path: str, target_dir: str, name: str) -> None:
        target_path = Path(target_dir).joinpath(name)
        if path.startswith("template:"):
            try:
                self._clone_catalog_template(path[len("template:"):], target_path, name)
            except (LookupError, ValueError):
                raise LookupError("Cannot resolve %s - make sure it's a valid name of a template located in %s!",
                                  (path, DefaultValues.catalog_template_url.value))
        else:
            self._clone_solution(path, target_path)

    def _clone_solution(self, path, target_path):
        """Copies a solution (by resolving and downloading) to a given target path."""
        resolve_result = self.album.collection_manager().resolve(path)
        target_path_solution = target_path.joinpath(DefaultValues.solution_default_name.value)
        file_operations.copy(resolve_result.path(), target_path_solution)
        module_logger().info('Copied solution %s to %s!' % (resolve_result.path(), target_path_solution))

    def _clone_catalog_template(self, template_name, target_path, name):
        """Clones a template by looking up the template name in the template catalog"""
        template_url = "%s/%s/-/archive/main/%s-main.zip" % (
            DefaultValues.catalog_template_url.value, template_name, template_name
        )
        if url_operations.is_downloadable(template_url):
            target_path.mkdir(parents=True)
            download_zip_target = self.album.configuration().cache_path_download().joinpath(template_name + ".zip")
            download_unzip_target = self.album.configuration().cache_path_download().joinpath(template_name)
            url_operations.download_resource(template_url, download_zip_target)
            file_operations.unzip_archive(download_zip_target, download_unzip_target)
            download_unzip_target_subdir = download_unzip_target.joinpath(f"{template_name}-main")
            file_operations.copy_folder(download_unzip_target_subdir, target_path, copy_root_folder=False)
            self.album.catalogs().create_new(target_path, name)
            module_logger().info('Downloaded template from %s to %s!' % (template_url, target_path))
            return True
        return False
